"""EZSpace provides the namespace object class for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.base import FastSpace
from worktoy.ezdata import IllegalInitException, IllegalMethodException
from worktoy.ezdata import NoDefaultError, AmbiguousDefaultError
from worktoy.ezdata import DefaultTypeMismatchError
from worktoy.meta import CallMeMaybe
from worktoy.parse import maybe
from worktoy.desc import AttriBox

try:
  from typing import Callable
except ImportError:
  Callable = object

try:
  from typing import TypeAlias
except ImportError:
  TypeAlias = object

try:
  Boxes: TypeAlias = list[tuple[str, AttriBox]]
except ImportError:
  Boxes = object
except TypeError:
  Boxes = object


class EZSpace(FastSpace):
  """Expands the FastSpace class with auto generated '__init__' method.
  Additionally, custom methods are disallowed. Classes that require
  methods should instead use FastObject. """

  __allow_init__ = False  # Allow the '__init__' method to be set.

  def _enableInit(self) -> None:
    """Enables the init method"""
    self.__allow_init__ = True

  def _disableInit(self, ) -> None:
    """Disables the init method"""
    self.__allow_init__ = False

  def _getAllowInit(self, ) -> bool:
    """Getter-function for allow init flag"""
    return True if self.__allow_init__ else False

  def __setitem__(self, key: str, value: object, **kwargs) -> None:
    """This method sets the key, value pair in the namespace."""
    if key == '__trust_me_bro__':
      if value:
        return self._enableInit()
      return self._disableInit()
    if kwargs.get('__trust_me_bro__', False):
      return FastSpace.__setitem__(self, key, value)
    if key == '__init__':
      if self._getAllowInit():
        self._disableInit()
        return FastSpace.__setitem__(self, key, value)
      clsName = self.getClassName()
      raise IllegalInitException(clsName, value)
    if self.isSpecialKey(key):
      return dict.__setitem__(self, key, value)
    if isinstance(value, AttriBox):
      return FastSpace.__setitem__(self, key, value)
    if callable(value):
      clsName = self.getClassName()
      raise IllegalMethodException(clsName, key, value)
    return FastSpace.__setitem__(self, key, value)

  def _unpackBoxes(self) -> tuple[list[str], list[object]]:
    """Unpacks the instances of AttriBox into keys and default values."""
    keys, defVals = [], []
    boxes = self._getAllBoxes()
    n = len(boxes)
    defVal = None
    clsName = self.getClassName()
    for (key, box) in boxes:
      keys.append(key)
      posArgs = AttriBox.getArgs(box, None, _root=True)
      if not posArgs:
        raise NoDefaultError(clsName, key, box)
      if len(posArgs) > 1:
        raise AmbiguousDefaultError(clsName, key, box)
      defVal = posArgs[0]
      fieldClass = AttriBox.getFieldClass(box)
      if not isinstance(defVal, fieldClass):
        raise DefaultTypeMismatchError(clsName, key, box, defVal)
      defVals.append(defVal)
    return keys, defVals

  def _getBoxKeys(self) -> list[str]:
    """Getter-function for the keys to the AttriBox instances."""
    return self._unpackBoxes()[0]

  def _getDefaultValues(self) -> list[object]:
    """Getter-function for the default values of the AttriBox instances."""
    return self._unpackBoxes()[1]

  def ezInitFactory(self, ) -> CallMeMaybe:
    """This factory creates a fast, but inflexible '__init__' method."""
    keys, defVals = self._unpackBoxes()
    n = len(keys)

    def __init__(instance, *args, ) -> None:
      pArgs = [*args, *[None, ] * (max([0, n - len(args)]))]
      for (k, arg, v) in zip(keys, pArgs, defVals):
        setattr(instance, k, maybe(arg, v))

    return __init__

  def ezStrFactory(self, ) -> CallMeMaybe:
    """Factory method for the '__str__' method."""
    keys, defVals = self._unpackBoxes()

    def __str__(instance, ) -> str:
      """String representation"""
      clsName = self.getClassName()
      valStr = [str(getattr(instance, key)) for key in keys]
      return '%s(%s)' % (clsName, ', '.join(valStr))

    return __str__

  def ezEqFactory(self) -> CallMeMaybe:
    """Factory method for the '__eq__' method"""
    keys, defVals = self._unpackBoxes()

    def __eq__(instance, other: object) -> bool:
      for key in keys:
        if getattr(instance, key) - getattr(other, key):
          return False
      else:
        return True

    return __eq__



"""The 'worktoy.info' module provides functionality relating to logging
and debugging. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._info_field import InfoField
from ._info_group import InfoGroup, BaseGroup
from ._info_meta import InfoMeta
from ._info_handle import InfoHandle



"""InfoGroup provides a grouping for info strings."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy._WORK_IN_PROGRESS.info import InfoMeta, InfoField

try:
  from typing import Callable, Any
except ImportError:
  Callable = object


class InfoGroup(metaclass=InfoMeta):
  """Info group categorizes the info strings."""

  name = InfoField('name', 'defaultGroupName')
  members = InfoField('members', [])

  def __init__(self, name: str) -> None:
    self.name = name

  def __str__(self, ) -> str:
    """String representation"""
    return """%s('%s')""" % (self.__class__.__name__, self.name)

  def __repr__(self, ) -> str:
    """String representation"""
    return """%s.%s""" % (self.__class__.__name__, self.name)


BaseGroup = InfoGroup('base')



"""TypeSig encapsulates type signatures and the functionality for
recognizing positional arguments. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any, Optional

from worktoy.text import monoSpace


class _SingleType:
  """This class specifies special type handling"""

  __inner_type__ = None

  def __init__(self, arg: object) -> None:
    """Initialize the _SingleType object."""
    if isinstance(arg, type):
      self.__inner_type__ = arg
    else:
      e = """The _SingleType must be initialized with a type, but received: 
      '%s'!"""
      raise TypeError(monoSpace(e % arg))

  def __str__(self) -> str:
    """String representation"""
    clsName = self.__inner_type__.__name__
    return '_SingleType(%s,)' % clsName

  def __repr__(self, ) -> str:
    """Code representation"""
    return self.__str__()

  def cast(self, arg: object) -> object:
    """Cast the argument to the inner type."""
    if arg is None:
      raise ValueError('None cannot be casted!')
    if isinstance(arg, self.__inner_type__):
      return arg
    if self.__inner_type__ is float:
      if isinstance(arg, int):
        return float(arg)
      if isinstance(arg, complex):
        if arg.imag:
          return None
        return float(arg.real)
    if self.__inner_type__ is int:
      if isinstance(arg, complex):
        if arg.imag:
          return None
        if arg.real.is_integer():
          return int(arg.real)
      if isinstance(arg, float):
        if arg.is_integer():
          return int(arg)
        return None
    return None


class TypeSig:
  """TypeSig encapsulates type signatures and the functionality for
  recognizing positional arguments. """

  __type_signature__ = None
  __single_types__ = None

  def __init__(self, *args) -> None:
    """Initialize the TypeSig object."""
    self.__type_signature__ = []
    self.__single_types__ = []
    for arg in args:
      if isinstance(arg, type):
        self.__type_signature__.append(arg)
        self.__single_types__.append(_SingleType(arg))
        continue
      e = """The TypeSig must be initialized with types, but received: 
      '%s'!"""
      raise TypeError(monoSpace(e % arg))

  def __contains__(self, other: tuple) -> bool:
    """Check if the TypeSig is contained in the other tuple."""
    if not other:
      return False
    if isinstance(other, (list, tuple)):
      out = self.cast(*other, )
      return False if out is None else True

  def __bool__(self, ) -> bool:
    """The empty signature makes the instance False."""
    return True if self.__type_signature__ else False

  def __len__(self) -> int:
    """Return the length of the type signature."""
    return len(self.__type_signature__)

  def cast(self, *args) -> Any:
    """Casts the arguments to the """
    if not args and not self:
      return []
    out = []
    for (type_, arg) in zip(self.__single_types__, args):
      val = _SingleType.cast(type_, arg)
      if val is None:
        return None
      out.append(val)
    return out

  def __str__(self) -> str:
    """String representation"""
    out = [cls.__name__ for cls in self.__type_signature__]
    return '(%s,)' % ', '.join(out)

  def __repr__(self) -> str:
    """String representation"""
    return self.__str__()



"""DEFAULT holds an object that disappears when accessed."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from typing import Any

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

from worktoy.meta import Zeroton


class DEFAULT(Zeroton):
  """DEFAULT is a class that holds an object that disappears when
  accessed."""

  __default_object__ = None
  __has_object__ = False

  def __class_call__(self, defaultObject: object) -> None:
    self.__default_object__ = defaultObject
    self.__has_object__ = True

  @classmethod
  def getDefaultObject(cls) -> Any:
    """Getter-function for the default object."""
    if cls.__has_object__:
      cls.__has_object__ = False
      return cls.__default_object__
    e = """The DEFAULT Zeroton holds no object!"""
    raise RuntimeError(e)

  if TYPE_CHECKING:
    def __init__(self, *args, **kwargs) -> None:
      """Trust me bro"""
      pass



"""The 'compileShellScript' function takes the path to a .env file,
reads the environment variables from the file, and creates a shell script
that checks if the environment variables are already set. If they are not
they are then set using the value from the file. The resulting shell
script is returned as a string including the shebang line such that both
bash and zsh should be able to run the script. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import time


def _getNotice() -> str:
  """Creates the notice that the shell script is auto-generated by zshpy. """
  autoGeneratedNotice = """This script was auto-generated by the zshpy 
  module. <br> Do not edit this file manually as it is managed 
  automatically meaning that changes will be overwritten at undefined 
  intervals. <br> This particular version of the script was generated on: %s
  """

  text = autoGeneratedNotice % '{{TIME}}'
  paragraphs = [line.strip() for line in text.split('<br>')]
  lines = []
  line = []
  lineLength = 77
  padLeft = '# %02d | '
  padRight = ' | #'
  nLeft = len(padLeft % 69)
  nRight = len(padRight)
  n = lineLength
  n0 = lineLength - nLeft - nRight
  for p in paragraphs:
    for word in p.split():
      if len(word) + len(' '.join(line)) > n0:
        lines.append(line)
        line = []
      line.append(word)
    lines.append(line)
    lines.append(' ')
    line = []
  lines.pop()
  fmtLines = []

  for (i, line) in enumerate(lines):
    left = padLeft % i
    mid = (' '.join(line)).ljust(n0, ' ')
    right = padRight
    fmtLines.append('%s%s%s' % (left, mid, right))
  header = ('_' * (n - 4)).center(n - 2, ' ').center(n, '#')
  footer = ('¨' * (n - 4)).center(n - 2, ' ').center(n, '#')
  out = [header, '\n'.join(fmtLines), footer]
  return '\n'.join(out).replace('{{TIME}}', '\n  ' + time.ctime() + '\n  ')


def compileShellScript(fid: str) -> str:
  """The 'compileShellScript' function takes the path to a .env file,
  reads the environment variables from the file, and creates a shell script
  that checks if the environment variables are already set. If they are not
  they are then set using the value from the file. The resulting shell
  script is returned as a string including the shebang line such that both
  bash and zsh should be able to run the script. """
  from moreworktoy import loadEnv

  envData = loadEnv(fid)

  script = "#!/bin/sh\n"
  #  Include comment explaining that the script is auto-generated by zshpy
  #  and that it should not be edited manually. Also, include a date for
  #  the generation of the script.

  for (key, value) in envData.items():
    script += f"if [ -z ${key} ]; then\n"
    script += f"  export {key}={value}\n"
    script += "fi\n"

  return script


def testLOL() -> str:
  """Testing _getNotice"""
  return _getNotice()



"""The 'applyEnv' function updates the environment with the variables in the
the given file."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from moreworktoy import loadEnv


def applyEnv(envFile: str) -> None:
  """The 'applyEnv' function updates the environment with the variables in
  the given file."""

  envData = loadEnv(envFile)

  for (key, value) in envData.items():
    os.environ[key] = value



"""This module provides custom exceptions for the 'worktoy.ezdata'
package."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.desc import AttriBox
from worktoy.meta import CallMeMaybe
from worktoy.text import monoSpace


class IllegalMethodException(TypeError):
  """This exception is raised when a subclass of EZData tries to implement
  a custom method. """

  def __init__(self, clsName: str, key: str, func: CallMeMaybe) -> None:
    e = """During class body execution for class: '%s', a custom method at 
    name: '%s' was encountered. EZData subclasses are not allowed to 
    implement methods! The function encountered was: '%s'."""
    e2 = e % (clsName, key, func)
    super().__init__(monoSpace(e2))


class IllegalInitException(IllegalMethodException):
  """This exception is raised when a subclass of EZData tries to implement
  a custom '__init__' method. """

  def __init__(self, clsName: str, func: object) -> None:
    e = """During class body execution for class: '%s', a custom '__init__'
    method was encountered. Subclasses are not allowed to implement the 
    '__init__' method! The function encountered was: %s""" % (clsName, func)
    TypeError().__init__(self, monoSpace(e))


class NoDefaultError(ValueError):
  """This exception is raised when an EZData subclass sets an AttriBox
  instance without a default value. """

  def __init__(self, clsName: str, key: str, box: AttriBox) -> None:
    e = """During class body execution for class '%s', an AttriBox 
    instance was encountered at name: '%s' having type: '%s', but no 
    default value could be found!"""
    e2 = e % (clsName, key, box)
    super().__init__(monoSpace(e2))


class AmbiguousDefaultError(ValueError):
  """This exception is raised when an EZData subclass sets an AttriBox
  instance with an ambiguous default value. """

  def __init__(self, clsName: str, key: str, box: AttriBox) -> None:
    e = """During class body execution for class '%s', an AttriBox 
    instance was encountered at name: '%s' having type: '%s', but the 
    default value is ambiguous!"""
    e2 = e % (clsName, key, box)
    super().__init__(monoSpace(e2))


class DefaultTypeMismatchError(TypeError):
  """This exception is raised when an EZData subclass sets an AttriBox
  instance with a default value of the wrong type. """

  def __init__(self,
               clsName: str,
               key: str,
               box: AttriBox,
               obj: object, ) -> None:
    attrName = key
    expType = AttriBox.getFieldClass(box).__name__
    actType = type(obj).__name__
    objStr = monoSpace(str(obj))
    e = """Attribute '%s' on class '%s' specifies type: '%s', but received 
    '%s' of type: '%s'!"""
    e2 = e % (attrName, clsName, expType, objStr, actType)
    super().__init__(monoSpace(e2))



"""The 'moreworktoy' module provides new stuff for worktoy created for use
by the current project, but being too general to be included in the
project."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._maybe import maybe
from ._maybe_type import maybeType



"""The 'validateFile' function receives a string and returns it if it
matches an existing file. If a second argument is given and the function
cannot validate the file, it returns the second argument otherwise the
error encountered is raised. This error may be suppressed by setting
keyword argument 'strict' to False.

For example, assuming some_file.txt is a file unlike sus_file.txt:
validateFile('some_file.txt') -> 'some_file.txt':
validateFile('sus_file.txt') -> Relevant error is raised:
validateFile('sus_file.txt', 'some_file.txt') -> 'some_file.txt':
validateFile('sus_file.txt', strict=False) -> '':

Passing both a second positional argument and setting 'strict' to False is
redundant."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os.path

from worktoy.text import typeMsg, monoSpace


def _parseFiles(*args) -> tuple:
  """The '_parseFiles' function receives a variable number of arguments and
  returns a tuple of the arguments that are strings. If no arguments are
  strings, an empty tuple is returned. """
  testFile, altFile = None, None
  for arg in args:
    if isinstance(arg, str):
      if testFile is None:
        testFile = arg
      else:
        altFile = arg
        break
    elif arg is not None:
      e = typeMsg('arg', arg, str)
      raise TypeError(e)
  return testFile, altFile


def validateFile(*args, **kwargs) -> str:
  """The 'validateFile' function receives a string and returns it if it
  matches an existing file. If a second argument is given and the function
  cannot validate the file, it returns the second argument otherwise the
  error encountered is raised. This error may be suppressed by setting
  keyword argument 'strict' to False.

  For example, assuming some_file.txt is a file unlike sus_file.txt:
  validateFile('some_file.txt') -> 'some_file.txt':
  validateFile('sus_file.txt') -> Relevant error is raised:
  validateFile('sus_file.txt', 'some_file.txt') -> 'some_file.txt':
  validateFile('sus_file.txt', strict=False) -> '':

  Passing both a second positional argument and setting 'strict' to False is
  redundant."""

  testFile, altFile = _parseFiles(*args)
  if testFile is None:
    e = """Received no file to validate!"""
    raise ValueError(e)
  if altFile is None and not kwargs.get('strict', True):
    altFile = ''

  if not os.path.exists(testFile):
    if altFile is not None:
      return altFile
    e = """Given file: '%s' does not exist!"""
    raise FileNotFoundError(monoSpace(e % testFile))
  if os.path.isdir(testFile):
    if altFile is not None:
      return altFile
    e = """Given file: '%s' is a directory!"""
    raise IsADirectoryError(monoSpace(e % testFile))
  try:
    with open(testFile, 'r') as f:
      f.read()
  except PermissionError as permissionError:
    if altFile is not None:
      return altFile
    e = """When trying to open file: '%s' for reading, encountered: %s"""
    e2 = e % (testFile, permissionError)
    raise PermissionError(monoSpace(e2)) from permissionError
  return testFile



"""The 'applyExistingEnv' function applies the existing environment to a
string containing references to environment variables. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os


def applyExistingEnv(s: str) -> str:
  """The 'applyExistingEnv' function applies the existing environment to a
  string containing references to environment variables. """
  for k, v in os.environ.items():
    s = s.replace(f"${k}", v)
  return s



"""The loadEnv function loads the environment from the file. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

import os

from worktoy.text import monoSpace

from moreworktoy import validateFile, findEnvNames


def loadEnv(fid: str) -> dict[str, str]:
  """The loadEnv function loads the environment from the file. """
  testFile = validateFile(fid)
  env = {}
  data = None

  with open(testFile, 'r') as f:
    data = f.readlines()

  if data is None:
    e = """Error reading file at: '%s'""" % testFile
    raise IOError(e)

  for line in data:
    if not line or line.startswith('#'):
      continue
    if '=' not in line:
      e = """The line: '%s' does not contain an equal sign!""" % line
      raise SyntaxError(monoSpace(e))
    entry = line.split('=')
    if len(entry) > 2:
      e = """The line: '%s' contains more than one equal sign!""" % line
      raise SyntaxError(monoSpace(e))
    key, value = entry
    for name in findEnvNames(value):
      value = value.replace('$%s' % name, os.environ.get(name, ''))
    env[key] = value.strip()
  return env



"""DoubleDollarException is raised when trying to find environment
variable names in a string containing two consecutive dollar signs. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace


class DoubleDollarException(SyntaxError):
  """DoubleDollarException is raised when trying to find environment
  variable names in a string containing two consecutive dollar signs. """

  def __init__(self, *args) -> None:
    if args:
      e = """The string: '%s' contains instances of consecutive dollar 
      signs which is not supported!""" % args[0]
      SyntaxError.__init__(self, monoSpace(e))
    else:
      SyntaxError.__init__(self, )



"""BaseNamespace provides the namespace object class for the
BaseMetaclass."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.meta import Overload

try:
  from typing import Callable, Any
except ImportError:
  Callable = object
  Any = object

from worktoy.meta import AbstractNamespace
from worktoy.parse import maybe
from worktoy.text import monoSpace

try:
  Overloaded = dict[tuple[type, ...], Callable]
except TypeError:
  Overloaded = dict


class OverloadEntry:
  """Instances of this class are used to store the overloads in the
  BaseNamespace."""

  __func_name__ = None
  __call_me_maybe__ = None
  __raw_types__ = None
  __assigned_key__ = None
  __nested_instances__ = None

  def __init__(self, *types) -> None:
    self.__raw_types__ = (*types,)

  def getNested(self) -> Any:
    """Getter-function for nested instances."""
    return maybe(self.__nested_instances__, [])

  def __call__(self, callMeMaybe: Any) -> Callable:
    existing = getattr(callMeMaybe, '__type_signatures__', [])
    self.__call_me_maybe__ = callMeMaybe
    self.__func_name__ = callMeMaybe.__name__
    setattr(callMeMaybe, '__type_signatures__', [*existing, self])
    return callMeMaybe

  def __str__(self, ) -> str:
    """String representation"""
    typeNames = [t.__name__ for t in self.__raw_types__]
    return """%s: %s""" % (self.__func_name__, typeNames)


class BaseNamespace(AbstractNamespace):
  """BaseNamespace provides the namespace object class for the
  BaseMetaclass."""

  __overload_dict__ = None

  def __init__(self, *args, **kwargs) -> None:
    """Initialize the BaseNamespace."""
    AbstractNamespace.__init__(self, *args, **kwargs)

  def __setitem__(self, key: str, value: object) -> None:
    """Set the item in the namespace."""
    overloadEntries = getattr(value, '__type_signatures__', None)
    if overloadEntries is None:
      return AbstractNamespace.__setitem__(self, key, value)
    existing = self.getOverloadList()
    for entry in overloadEntries:
      entry.__assigned_key__ = key
    self.__overload_dict__ = [*existing, *overloadEntries]

  def getOverloadList(self) -> list:
    """Getter-function for overloads"""
    return maybe(self.__overload_dict__, [])

  def compile(self) -> dict:
    """Compile the namespace into a dictionary."""
    out = {}
    for entry in self.getOverloadList():
      name = entry.__func_name__
      key = entry.__assigned_key__
      sig = entry.__raw_types__
      func = entry.__call_me_maybe__
      if name != key:
        e = """ERROR IN COMPILE!
        The function name '%s' does not match the assigned key 
        '%s'! END"""
        raise ValueError(monoSpace(e % (name, key)))
      overloadField = out.get(name, Overload())
      overloadField.overload(*sig)(func)

      out[name] = overloadField

    return {**AbstractNamespace.compile(self), **out}



"""The 'layer' module provides functionality relating to loading data and
environment variables."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from ._double_dollar_exception import DoubleDollarException
from ._validate_file import validateFile
from ._apply_existing_env import applyExistingEnv
from ._find_env_names import findEnvNames
from ._load_env import loadEnv
from ._apply_env import applyEnv



"""This method finds environment variable names in a given string. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from string import ascii_letters, digits

from moreworktoy import DoubleDollarException


def findEnvNames(source: str) -> list[str]:
  """This method finds environment variable names in a given string. """
  if '$' not in source:
    return []
  if '$$' in source:
    raise DoubleDollarException(source)
  n = len(source)
  nameChars = ascii_letters + digits + '_'
  names = []
  for (i, char) in enumerate(source):
    if char == '$':
      j = 0
      while i + j + 1 < n and source[i + j + 1] in nameChars:
        j += 1
      if j > 0:
        name = source[i + 1:i + j + 1]
        if name[0] in digits:
          continue
        return [name, *findEnvNames(source[i + j + 1:])]
  return []



"""AttriBox provides a feature complete implementation of the descriptor
protocol with lazy instantiation. With it, the owning class need only set
one attribute on one line to access the full functionality of the
descriptor. The syntactic sugar is as follows:


class Info:
  # This class provides an attribute class of the Owner class below

  __owning_instance__ = None

  def __init__(self, instance: object) -> None:
    self.__owning_instance__ = instance


class Owner:
  # This class owns attributes through the AttriBox class

  x = AttriBox[float]()
  info = AttriBox[Info](THIS)  # THIS is replaced by the owning instance.


The class of the attribute is placed in the brackets and the parentheses
are given the arguments used to instantiate the attribute. It is possible
to pass special placeholders here which are replaced when the attribute
object is created. The placeholders are:

THIS: The owning instance
TYPE: The owning class
BOX: The AttriBox instance
ATTR: The attribute class or its subclass

The lifecycle of the AttriBox instance is as follows:

1. The AttriBox class itself is created
2. The AttriBox instance is created during the class body execution of a
class that is being created.
3. When the class creation process completes, the '__set_name__' method is
invoked. This class is inherited from the 'CoreDescriptor' class.
4. When this AttriBox instance is accessed through the owning class,
not an instance of it, the AttriBox instance itself returns.
5. When the access is through an instance of the owning class,
the AttriBox instance first attempts to find an existing value in the
namespace of the instance at its private name. This value is returned if
it exists.
6. Otherwise, an instance of the wrapping class 'Bag' is created and an
instance of the inner class is created and stored in the 'Bag' instance.
It is the 'Bag' instance that is stored in the namespace of the owning
class and during calls to __get__, the wrapped object is returned from
inside the Bag instance.
7. By default, the setter method expects the value received to be of the
same type as the existing object in the Bag instance.
8. By default, the deleter method is disabled and will raise an exception.
This is because calls on the form: 'del instance.attribute' must then
cause 'instance.attribute' to result in Attribute error. This is not
really practical as it is insufficient to remove the value as the
descriptor is on the owning class not the instance. This means that no
functionality is present to distinguish between the __delete__ having been
called, and then inner object not having been created yet.

"""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Self
except ImportError:
  Self = object

try:
  from typing import Never
except ImportError:
  Never = object

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

from typing import Any

from worktoy.desc import Bag, THIS, TYPE, AbstractDescriptor, BOX, \
  ATTR, DEFAULT, NODEF
from worktoy.parse import maybe
from worktoy.text import typeMsg, monoSpace


class AttriBox(AbstractDescriptor):
  """AttriBox class improves the AttriBox class!"""

  __default_object__ = None
  __field_class__ = None
  __pos_args__ = None
  __key_args__ = None

  @staticmethod
  def _castNumberTo(value: object, target: type) -> Any:
    """This method handles the cringe cases of ints being rejected as
    floats. """
    if isinstance(value, int):
      if target is int:
        return value
      if target is float:
        return float(value)
      if target is complex:
        return float(value) + 0j
    if isinstance(value, float):
      if target is int:
        if value.is_integer():
          return int(value)
        e = typeMsg('value', value, int)
        raise TypeError(e)
      if target is float:
        return value
      if target is complex:
        return value + 0j
    if isinstance(value, complex):
      if target is int:
        if value.real.is_integer() and not value.imag:
          return int(value.real)
        e = typeMsg('value', value, int)
        raise TypeError(e)
      if target is float:
        if not value.imag:
          return float(value.real)
        e = typeMsg('value', value, float)
        raise TypeError(e)
      if target is complex:
        return value

  @staticmethod
  def _unpackValueToArgsKwargs(value: Any) -> tuple[tuple, dict]:
    """This method unpacks the value into positional and keyword
    arguments."""
    if not value:
      return (), {}
    if not isinstance(value, (list, tuple)):
      return (value,), {}
    if len(value) == 1:
      return (value[0],), {}
    if len(value) == 2:
      if isinstance(value[0], (list, tuple)):
        if isinstance(value[1], dict):
          return (*value[0],), {**value[1], }
    return (*value,), {}

  @classmethod
  def __class_getitem__(cls, fieldClass: type) -> AttriBox:
    """Class method for creating a AttriBox instance."""
    return cls(fieldClass)

  def __init__(self, *args) -> None:
    AbstractDescriptor.__init__(self)
    for arg in args:
      if isinstance(arg, type):
        fieldClass = arg
        break
    else:
      e = """AttriBox constructor requires the fieldClass type!"""
      raise ValueError(e)
    if isinstance(fieldClass, type):
      self.__field_class__ = fieldClass
    else:
      e = """AttriBox constructor requires the fieldClass type!"""
      e2 = typeMsg('fieldClass', fieldClass, type)
      raise TypeError(monoSpace('%s\n%s' % (e, e2)))
    self.__field_class__ = fieldClass

  def __call__(self, *args, **kwargs) -> Self:
    for arg in args:
      if arg is DEFAULT:
        self.__default_object__ = DEFAULT.getDefaultObject()
        if not isinstance(self.__default_object__, self.getFieldClass()):
          e = """The default object is not of the correct type!"""
          e2 = typeMsg('defaultObject', self.__default_object__,
                       self.getFieldClass())
          raise TypeError(monoSpace('%s\n%s' % (e, e2)))
        break
      if arg is NODEF:
        if len(args) - 1 or kwargs:
          e = """The NODEF flag must be the only argument!"""
          raise ValueError(e)
        self.__default_object__ = NODEF
        break
    else:
      self.__pos_args__ = [*args, ]
      self.__key_args__ = {**kwargs, }
    return self

  def getFieldClass(self, ) -> type:
    """Getter-function for the field class."""
    if self.__field_class__ is None:
      e = """The field class of the AttriBox instance has not been set!"""
      raise AttributeError(e)
    if isinstance(self.__field_class__, type):
      return self.__field_class__
    e = typeMsg('__field_class__', self.__field_class__, type)
    raise TypeError(e)

  def getArgs(self, instance: object, **kwargs) -> list[Any]:
    """Getter-function for positional arguments"""
    if kwargs.get('_root', False):
      return maybe(self.__pos_args__, [])
    args = []
    for arg in maybe(self.__pos_args__, []):
      if arg is THIS:
        args.append(instance)
      elif arg is TYPE:
        args.append(self.getFieldOwner())
      elif arg is BOX:
        args.append(self)
      elif arg is ATTR:
        args.append(type(self))
      else:
        args.append(arg)
    return args

  def getKwargs(self, instance: object, **kwargs) -> dict[str, Any]:
    """Getter-function for keyword arguments"""
    if kwargs.get('_root', False):
      return maybe(self.__key_args__, {})
    keyArgs = {}
    for (key, val) in maybe(self.__key_args__, {}).items():
      if val is THIS:
        keyArgs[key] = instance
      elif val is TYPE:
        keyArgs[key] = self.getFieldOwner()
      elif val is BOX:
        keyArgs[key] = self
      elif val is ATTR:
        keyArgs[key] = type(self)
      else:
        keyArgs[key] = val
    return keyArgs

  def getDefaultFactory(self) -> Any:
    """Getter-function for function creating the default value. """
    keyArgs = self.getKwargs(None, _root=True)
    posArgs = self.getArgs(None, _root=True)
    fieldClass = self.getFieldClass()

    zerotons = [BOX, ATTR]
    for zero in zerotons:
      if zero in [*posArgs, *keyArgs.values()]:
        e = """The 'getDefaultFactory' does not BOX and ATTR, 
        but received: %s""" % zero
        raise ValueError(monoSpace(e))

    def callMeMaybe(instance: object) -> Any:
      """This function creates the default value."""
      newKeys = {}
      for (key, val) in keyArgs.items():
        if val is THIS:
          newKeys[key] = instance
        elif val is TYPE:
          newKeys[key] = type(instance)
        else:
          newKeys[key] = val
      newArgs = []
      for arg in posArgs:
        if arg is THIS:
          newArgs.append(instance)
        elif arg is TYPE:
          newArgs.append(type(instance))
        else:
          newArgs.append(arg)
      if fieldClass is bool:
        innerObject = True if [*newArgs, None][0] else False
      else:
        innerObject = fieldClass(*newArgs, **newKeys)
      if TYPE_CHECKING:
        return Bag(None, innerObject)
      return innerObject

    return callMeMaybe

  def createFieldObject(self, instance: object, ) -> Bag:
    """Create the field object. If the default object is set, it is used."""
    if self.__default_object__ is not None:
      return Bag(instance, self.__default_object__)
    if self.__field_class__ is None:
      e = """AttriBox instance has not been initialized!"""
      raise AttributeError(e)
    if self.__pos_args__ is None or self.__key_args__ is None:
      e = """AttriBox instance has not been called!"""
      raise AttributeError(e)
    keyArgs = self.getKwargs(instance, )
    posArgs = self.getArgs(instance, )
    fieldClass = self.getFieldClass()
    if fieldClass is bool:
      innerObject = True if [*posArgs, None][0] else False
    else:
      innerObject = fieldClass(*posArgs, **keyArgs)
    return Bag(instance, innerObject)

  def __instance_reset__(self, instance: object) -> None:
    """Reset-function for the instance."""
    pvtName = self._getPrivateName()
    delattr(instance, pvtName)

  def __instance_get__(self, instance: object, **kwargs) -> Any:
    """Getter-function for the instance. Please note, that if the call is
    the notifier asking what the previous value was, the functionality in
    the AbstractDescriptor expects and handles the exception. """
    pvtName = self._getPrivateName()
    if getattr(instance, pvtName, None) is None:
      if self.__default_object__ is NODEF:
        e = """This instance of %s has the NODEF flag set, and the 
        required explicitly set value is missing! NODEF requires that an 
        explicit value be set for the attribute.""" % self.getFieldName()
        raise TypeError(monoSpace(e))
      if kwargs.get('_recursion', False):
        raise RecursionError
      setattr(instance, pvtName, self.createFieldObject(instance))
      return self.__instance_get__(instance, _recursion=True)
    bag = getattr(instance, pvtName)
    if not isinstance(bag, Bag):
      e = typeMsg('bag', bag, Bag)
      raise TypeError(e)
    innerObject = bag.getInnerObject()
    fieldClass = self.getFieldClass()
    if isinstance(innerObject, fieldClass):
      return innerObject
    e = typeMsg('innerObject', innerObject, fieldClass)
    raise TypeError(e)

  def __instance_set__(self,
                       instance: object,
                       value: object,
                       **kwargs) -> None:
    """Setter-function for the instance."""
    pvtName = self._getPrivateName()
    fieldCls = self.getFieldClass()
    if isinstance(value, fieldCls):
      bag = getattr(instance, pvtName, None)
      if bag is None:
        return setattr(instance, pvtName, Bag(instance, value))
      bag.setInnerObject(value)
      return setattr(instance, pvtName, bag)
    if fieldCls in [int, float, complex]:
      if kwargs.get('_recursion2', False):
        e = typeMsg('value', value, fieldCls)
        raise TypeError(e)
      return self.__instance_set__(instance,
                                   self._castNumberTo(value, fieldCls),
                                   _recursion2=True)
    if not isinstance(value, (tuple, list)):
      if kwargs.get('_recursion3', False):
        raise RecursionError
      return self.__instance_set__(instance, (value,), _recursion3=True)
    if kwargs.get('_recursion', False):
      e = typeMsg('value', value, fieldCls)
      raise TypeError(e)
    newValue = None
    try:
      args, kwargs2 = self._unpackValueToArgsKwargs(value)
      newValue = fieldCls(*args, **kwargs2)
    except ValueError as valueError:
      if 'could not convert' in str(valueError):
        e = """Could not convert the value to the field class!"""
        e2 = typeMsg('value', value, fieldCls)
        raise TypeError(monoSpace('%s\n%s' % (e, e2)))
    return self.__instance_set__(instance, newValue, _recursion=True)

  def __instance_del__(self, instance: object) -> Never:
    """Deleter-function for the instance."""
    e = """Deleter-function is not implemented by the AttriBox class."""
    raise TypeError(e)

  def __str__(self, ) -> str:
    """String representation"""
    posArgs = self.getArgs(None, _root=True)
    keyArgs = self.getKwargs(None, _root=True)
    posStr = ', '.join([str(arg) for arg in posArgs])
    keyStr = ', '.join([f'{k}={v}' for (k, v) in keyArgs.items()])
    argStr = ', '.join([arg for arg in [posStr, keyStr] if arg])
    clsName = self.getFieldClass().__name__
    return """AttriBox[%s](%s)""" % (clsName, argStr)

  def __repr__(self, ) -> str:
    """String representation"""
    return str(self)



"""InfoField provides a primitive implementation of the descriptor
protocol used by the rest of the 'worktoy.info' package. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace
from typing import Any

try:
  from typing import Callable
except ImportError:
  Callable = object


class InfoField:
  """Descriptor for the info"""

  __field_name__ = None
  __field_owner__ = None
  __pvt_name__ = None
  __default_value__ = None

  def _getPrivateName(self) -> str:
    """Getter-function for the private name"""
    if self.__pvt_name__ is None:
      e = """The field name is not set!"""
      raise ValueError(monoSpace(e))
    return self.__pvt_name__

  def _getDefaultValue(self) -> str:
    """Getter-function for the default value!"""
    if self.__default_value__ is None:
      e = """The default value is not set!"""
      raise ValueError(monoSpace(e))
    return self.__default_value__

  def __set_name__(self, owner: type, name: str) -> None:
    """Set the name of the field and the owner of the field."""
    self.__field_name__ = name
    self.__field_owner__ = owner
    if self.__pvt_name__ is None:
      self.__pvt_name__ = '_%s' % self.__field_name__

  def __init__(self, *args, **kwargs) -> None:
    """The '__init__' method initializes the descriptor."""
    posArgs = [*args, None, None][:2]
    self.__default_value__, self.__pvt_name__ = posArgs

  def __get__(self, instance: object, owner: type, **kwargs) -> Any:
    """Getter-function"""
    if instance is None:
      return self
    pvtName = self._getPrivateName()
    value = getattr(instance, pvtName, None)
    if value is None:
      if kwargs.get('_recursion', False):
        raise RecursionError
      self.__set__(instance, self._getDefaultValue())
      return self.__get__(instance, owner, _recursion=True)
    return value

  def __set__(self, instance: object, value: object) -> None:
    """Setter-function"""
    setattr(instance, self._getPrivateName(), value)



"""DispatchException provides a custom exception raised when an instance
of OverloadDispatcher fails to resolve the correct function from the
given arguments. Because the overload protocol relies on type matching,
this exception subclasses TypeError such that it can be caught by external
error handlers. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.text import monoSpace

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

if TYPE_CHECKING:
  from worktoy.meta import Overload


class DispatchException(TypeError):
  """DispatchException provides a custom exception raised when an instance
  of OverloadDispatcher fails to resolve the correct function from the
  given arguments. """

  __overload_function__ = None
  __pos_args__ = None

  def __init__(self, func: Overload, *args) -> None:
    e = """When calling the overloaded function '%s' the dispatcher could 
    not resolve the correct function from the given arguments: %s!"""
    self.__overload_function__ = func
    self.__pos_args__ = [*args, ]
    argStr = ["""  '%s' of type '%s'""" % (arg, type(arg)) for arg in args]
    argStr = ',\n'.join(argStr)
    errorMsg = monoSpace(e % (func.__function_name__, argStr))
    TypeError.__init__(self, errorMsg)

  def getOverload(self) -> Overload:
    """Return the Overload instance."""
    return self.__overload_function__

  def getArgs(self) -> list:
    """Return the arguments causing the exception."""
    return self.__pos_args__



"""The EZMeta class provides the metaclass for the EZData class."""
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy.ezdata import EZSpace
from worktoy.meta import BaseMetaclass, Bases


class EZMeta(BaseMetaclass):
  """The EZMeta class provides the metaclass for the EZData class."""

  @classmethod
  def __prepare__(mcls, name: str, bases: Bases, **kwargs) -> EZSpace:
    return EZSpace(mcls, name, bases, **kwargs)

  def __new__(mcls,
              name: str,
              bases: Bases,
              space: EZSpace,
              **kwargs) -> type:
    newInit = space.ezInitFactory()
    newStr = space.ezStrFactory()
    newEq = space.ezEqFactory()
    EZSpace.__setitem__(space, '__init__', newInit, __trust_me_bro__=True)
    EZSpace.__setitem__(space, '__str__', newStr, __trust_me_bro__=True)
    EZSpace.__setitem__(space, '__repr__', newStr, __trust_me_bro__=True)
    EZSpace.__setitem__(space, '__eq__', newEq, __trust_me_bro__=True)
    return BaseMetaclass.__new__(mcls, name, bases, space, **kwargs)



"""InfoMeta provides a needlessly complicated way of logging stuff.
Allegedly. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

try:
  from typing import Callable
except ImportError:
  Callable = object

try:
  from typing import TYPE_CHECKING
except ImportError:
  TYPE_CHECKING = False

try:
  from typing import Type
except ImportError:
  if TYPE_CHECKING:
    Type = type('Type', (), dict(__getitem__=lambda cls, key: cls.__class__))
  else:
    Type = {object: type}

from worktoy.parse import maybe


class InfoMeta(type):
  """Sets descriptor on class level"""

  __cls_instances__ = None

  def _getInstances(cls) -> dict:
    """Getter-function for list of instances"""
    return maybe(cls.__cls_instances__, {})

  def _addInstance(cls, self: object) -> None:
    """Setter-function for list of instances"""
    instances = cls._getInstances()
    cls.__cls_instances__ = {**instances, getattr(self, 'name'): self}

  def __getattr__(cls, key: str) -> object:
    """Allows the dot to return a name instance. """
    value = cls._getInstances().get(key, None)
    if value is None:
      return object.__getattribute__(cls, key)

  def __call__(cls, name: str) -> object:
    """The '__call__' method is invoked when the class is called."""

    self = type.__call__(cls, name)
    cls._addInstance(self)

    return self



"""InfoHandle encapsulates messages along with an InfoGroup permitting
grouping of messages. """
#  AGPL-3.0 license
#  Copyright (c) 2024 Asger Jon Vistisen
from __future__ import annotations

from worktoy._WORK_IN_PROGRESS.info import InfoGroup, InfoField, BaseGroup
from worktoy.text import typeMsg

try:
  from typing import Callable, Any
except ImportError:
  Callable = object

from worktoy.parse import maybe


class InfoHandle:
  """LOL"""

  __info_instances__ = None

  @classmethod
  def _getInfoInstances(cls) -> list[InfoHandle]:
    """Getter-function for list of instances"""
    return maybe(cls.__info_instances__, [])

  @classmethod
  def _registerInfoInstance(cls, self: InfoHandle) -> None:
    """Registers the instance"""
    instances = cls._getInfoInstances()
    if self not in instances:
      cls.__info_instances__ = [*instances, self]

  @classmethod
  def _getGroupedInstances(cls, infoGroup: InfoGroup) -> list[InfoHandle]:
    """Returns the instances belonging to the given group"""
    print('_getGroupedInstances: ', infoGroup)
    base = cls._getInfoInstances()
    print('\n'.join([str(i) for i in base]))
    out = [i for i in base if i.grp == infoGroup]
    print(out)
    return out

  msg = InfoField('')
  grp = InfoField(BaseGroup)

  def __init__(self, *args) -> None:
    """The '__init__' method initializes the descriptor."""
    if not args:
      e = """The message is not set!"""
      raise ValueError(e)
    if len(args) == 1:
      self.msg = args[0]
    elif len(args) > 1:
      self.msg, self.grp = [*args, ][:2]
    self._registerInfoInstance(self)
    if self not in self._getInfoInstances():
      raise ValueError
    print("""lmao: %s = %s""" % (str(self.grp), self.msg))

  def __matmul__(self, other: object) -> InfoHandle:
    """The '__matmul__' method is invoked when the '@' operator is used."""
    if isinstance(other, InfoGroup):
      print(self)
      self.grp = other
      print(self)
      return self
    if isinstance(other, str):
      return self @ InfoGroup(other)
    return NotImplemented

  def __str__(self, ) -> str:
    """String representation"""
    return """InfoHandle('%s'): \n--<| %s""" % (str(self.grp), self.msg)

  def __repr__(self, ) -> str:
    """String representation"""
    return """InfoHandle('%s', %s)""" % (self.msg, self.grp)

  def __class_getitem__(cls, infoGroup: object, **kwargs) -> None:
    """The '__class_getitem__' method is invoked when the class is
    indexed."""
    if isinstance(infoGroup, InfoGroup):
      header = """Beginning Printout of group: '%s'""" % str(infoGroup)
      body = []
      for self in cls._getGroupedInstances(infoGroup):
        body.append(str(self))
      else:
        footer = """End of Printout"""
        lines = [header, *body, footer]
        n = max([len(i) for i in lines])
        header = header.center(n, '_').center(n + 4, '|')
        footer = footer.center(n, '¨').center(n + 4, '|')
        lines = [line.ljust(n, ' ').center(n + 4, '|') for line in
                 lines[1:-1]]
        print(header)
        for line in lines:
          print(line)
        print(footer)
        return
    elif isinstance(infoGroup, str):
      if kwargs.get('_recursion', False):
        raise RecursionError
      grp = getattr(InfoGroup, infoGroup, None)
      if grp is None:
        return cls.__class_getitem__(InfoGroup(infoGroup), _recursion=True)
      return cls.__class_getitem__(infoGroup, _recursion=True)
    e = typeMsg('infoGroup', infoGroup, InfoGroup)
    raise TypeError(e)
